import { groupBy, map, mapObjIndexed, partition, pipe, sum } from "ramda";
import { StrictOmit } from "ts-essentials";
import { UploadStatisticsMimeRecord } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeRecord";
import { StorageStatisticsSegmentDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsSegmentDto";
import { UploadStatisticsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsDto";
import { UploadStatisticsMimeSegment } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeSegment";
import { TypedFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/TypedFileEvent";
import { BigSizeInBytes, EpochMillis, FileTagName } from "@upload-io/upload-shared-iso/common/model/Primitives";
import { Milliseconds, SizeInBytes } from "@upload-io/upload-shared-iso-lite";
import { isNonEmptyArray, NonEmpty, NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { BasicFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/BasicFileEvent";
import { IsoTimeUtils } from "@upload-io/upload-shared-iso/common/IsoTimeUtils";
import { BoxedArithmetic } from "@upload-io/upload-shared-iso/common/BoxedArithmetic";
import { BytesByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BytesByEpochDto";
import { DailyMonthlyBytesDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyBytesDto";
import { StorageStatisticsByTimeDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsByTimeDto";
import { InvocationEvent } from "@upload-io/upload-shared-iso/statistics/storage/InvocationEvent";
import { DurationByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DurationByEpochDto";
import { DailyMonthlyDurationsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyDurationsDto";
import { StorageStatisticsBySourceDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsBySourceDto";
import { DailyMonthlyUploadsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyUploadsDto";
import { UploadsByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadsByEpochDto";
import { UploadsAtEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadsAtEpochDto";

const daysToRecord = 60; // UI requires at least 30 for 'last 30 days' widget. We capture more for good measure.
const monthsToRecord = 24; // UI requires at least 12 for 'monthly bandwidth' widget. We capture more for good measure.

type Predicate<T> = (item: T) => boolean;
type MimeFilters = StrictOmit<UploadStatisticsMimeRecord<Predicate<TypedFileEvent>>, "other">;
type EventsByMime = UploadStatisticsMimeRecord<TypedFileEvent[]>;
type StatisticsByMime = UploadStatisticsMimeRecord<StorageStatisticsSegmentDto>;
type EventsFunc<T> = (events: TypedFileEvent[]) => T;

const mimeFilters: MimeFilters = {
  audio: x => x.mime.startsWith("audio/"),
  image: x => x.mime.startsWith("image/"),
  video: x => x.mime.startsWith("video/"),
  document: x =>
    [
      "application/x-abiword",
      "application/msword",
      "application/vnd.ms-",
      "application/vnd.openxmlformats",
      "application/vnd.oasis.opendocument",
      "application/pdf"
    ].some(prefix => x.mime.startsWith(prefix))
};

const mimeFiltersEntries: Array<[UploadStatisticsMimeSegment, Predicate<TypedFileEvent>]> = Object.entries(
  mimeFilters
).map(([k, v]) => [k as UploadStatisticsMimeSegment, v]);

export class StatisticsCalculator {
  private static readonly dashboardUploadTag = FileTagName("via/manual_upload"); // Sync with: upload-website

  static calculateStatistics(
    newEvents: NonEmptyArray<TypedFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): UploadStatisticsDto {
    const groupByMime: EventsFunc<EventsByMime> = groupBy(event => {
      const match = mimeFiltersEntries.find(([_, filter]) => filter(event));
      if (match !== undefined) {
        return match[0];
      }
      return "other";
    });

    const calculateStatisticsForMime = map<EventsByMime, StatisticsByMime>(
      (events: TypedFileEvent[]) => this.calculateStatisticsSegment(NonEmpty(events), sizeMultiplier, countMultiplier) // Input is from a groupBy, so we know is non-empty.
    );

    const newStatisticsMime: Partial<StatisticsByMime> = pipe(groupByMime, calculateStatisticsForMime)(newEvents);

    return {
      all: this.calculateStatisticsSegment(newEvents, sizeMultiplier, countMultiplier),
      byMime: {
        ...UploadStatisticsDto.empty.byMime,
        ...newStatisticsMime
      },
      bySource: this.calculateStorageStatisticsBySource(newEvents, sizeMultiplier, countMultiplier),
      byTime: this.calculateStorageStatisticsByTime(newEvents, sizeMultiplier, countMultiplier)
    };
  }

  static calculateMonthlyDailyBytes(events: BasicFileEvent[]): DailyMonthlyBytesDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyUploads(events: BasicFileEvent[]): DailyMonthlyUploadsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateUploadsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateUploadsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyDurations(events: InvocationEvent[]): DailyMonthlyDurationsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateStatisticsSegment(
    newEvents: NonEmptyArray<BasicFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): StorageStatisticsSegmentDto {
    const timestamps = newEvents.map(x => x.timeOfEvent);
    const eventSizeSum = sum(newEvents.map(x => x.size));
    const eventCount = newEvents.length;
    return {
      firstFileCreated: sizeMultiplier === -1 ? undefined : EpochMillis(Math.min(...timestamps)),
      lastFileCreated: sizeMultiplier === -1 ? undefined : EpochMillis(Math.max(...timestamps)),
      totalFileCount: countMultiplier * eventCount,
      totalFileCountMonotonic: Math.max(0, countMultiplier) * eventCount,
      totalFileSize: SizeInBytes(sizeMultiplier * eventSizeSum), // Can go negative (in a shard)
      totalFileSizeMonotonic: SizeInBytes(Math.max(0, sizeMultiplier) * eventSizeSum)
    };
  }

  static calculateStorageStatisticsBySource(
    newEvents: NonEmptyArray<TypedFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): StorageStatisticsBySourceDto {
    const [dashboard, other] = partition(x => this.isFileFromDashboard(x.tags), newEvents);
    return {
      dashboard: isNonEmptyArray(dashboard)
        ? this.calculateStatisticsSegment(dashboard, sizeMultiplier, countMultiplier)
        : StorageStatisticsSegmentDto.empty,
      other: isNonEmptyArray(other)
        ? this.calculateStatisticsSegment(other, sizeMultiplier, countMultiplier)
        : StorageStatisticsSegmentDto.empty
    };
  }

  static isFileFromDashboard(tags: FileTagName[]): boolean {
    return tags.some(x => x === this.dashboardUploadTag);
  }

  static calculateStorageStatisticsByTime(
    newEvents: NonEmptyArray<BasicFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): StorageStatisticsByTimeDto {
    return {
      creates: sizeMultiplier === 1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty,
      createsExtended:
        countMultiplier === 1 ? this.calculateMonthlyDailyUploads(newEvents) : DailyMonthlyUploadsDto.empty,
      deletes: sizeMultiplier === -1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty
    };
  }

  private static calculateBytesPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): BytesByEpochDto {
    const totalSize = (events: BasicFileEvent[]): BigSizeInBytes =>
      BoxedArithmetic.bigSizeInBytes.sum(events.map(event => BigSizeInBytes(event.size.toString())));

    return StatisticsCalculator.reducePerEpochUnit(getEpochUnit, keepLastNUnits, events, now, totalSize);
  }

  private static calculateUploadsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): UploadsByEpochDto {
    const totalCount = (events: BasicFileEvent[]): UploadsAtEpochDto => ({
      totalCount: events.length
    });

    return StatisticsCalculator.reducePerEpochUnit(getEpochUnit, keepLastNUnits, events, now, totalCount);
  }

  private static reducePerEpochUnit<T>(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis,
    reduce: (group: BasicFileEvent[]) => T
  ): {
    [epochValue: string]: T | undefined;
  } {
    const eventsByEpoch = groupBy<BasicFileEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(reduce, eventsByEpoch));
  }

  private static calculateDurationsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: InvocationEvent[],
    now: EpochMillis
  ): DurationByEpochDto {
    const eventsByEpoch = groupBy<InvocationEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: InvocationEvent[]): Milliseconds =>
      Milliseconds(sum(events.map(event => event.duration)));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }
}
